"""
From http://en.wikipedia.org/wiki/Scalpel:

    A scalpel, or lancet, is a small and extremely sharp bladed instrument used
    for surgery, anatomical dissection, and various arts and crafts (called a
    hobby knife).
"""

__author__ = 'Jonathan Stoppani'
__email__ = 'jonathan@stoppani.name'
__version__ = '0.6.0'
__url__ = 'https://github.com/GaretJax/lancet'
